package project5

import util._

/**
 * Second Part
 */

trait Constraint {
  def newValue: Unit
  def dropValue: Unit
}

object NoConstraint extends Constraint {
  def newValue { sys.error("NoConstraint.newValue") }
  def dropValue { sys.error("NoConstraint.dropValue") }
}

case class Adder(a1: Quantity, a2: Quantity, sum: Quantity) extends Constraint {
  def newValue {
    (a1.getValue, a2.getValue, sum.getValue) match {
    case (Some(x1), Some(x2), _      ) => sum.setValue(x1 + x2, this)
    case (Some(x1), _       , Some(r)) => a2.setValue(r - x1, this)
    case (_       , Some(x2), Some(r)) => a1.setValue(r - x2, this)
    case _                             =>
    }
  }
  def dropValue {
    a1.forgetValue(this); a2.forgetValue(this); sum.forgetValue(this)
  }
  a1 connect this
  a2 connect this
  sum connect this
}

case class Constant(q: Quantity, v: Double) extends Constraint {
  def newValue { sys.error("Constant.newValue") }
  def dropValue { sys.error("Constant.dropValue") }
  q connect this
  q.setValue(v, this)
}

case class Probe(name: String, q: Quantity) extends Constraint {
  def newValue { printProbe(q.getValue) }
  def dropValue { printProbe(None) }
  private def printProbe(v: Option[Double]) {
    val vstr = v match {
    case Some(x) => x.toString()
    case None => "?"
    }
    println("Probe: " + name + " = " + vstr)
  }
  q connect this
}


/**
 * Exercise 1
 */

// Complete with the constraints for multiplication and equality

// case class Multiplier ...

// case class Eq ...

/**
 * Exercise 2
 */

// Complete with the constraint for square

// case class Squarer ...


class Quantity {

  private var value: Option[Double] = None
  private var constraints: List[Constraint] = List()
  private var informant: Constraint = null

  def getValue: Option[Double] = value

  def setValue(v: Double, setter: Constraint) {
    value match {
    case Some(v1) =>
      if (v != v1) sys.error("Error! contradiction: " + v + " and " + v1)
    case None =>
      informant = setter
      value = Some(v)
      for (c <- constraints if c != informant) c.newValue
    }
  }
  def setValue(v: Double) { setValue(v, NoConstraint) }

  def forgetValue(retractor: Constraint) {
    if (retractor == informant) {
      value = None
      for (c <- constraints if c != informant) c.dropValue
    }
  }
  def forgetValue { forgetValue(NoConstraint) }

  def connect(c: Constraint) {
    constraints = c :: constraints
    value match {
      case Some(_) => c.newValue
      case None    =>
    }
  }

  def +(that: Quantity): Quantity = {
    val sum = new Quantity
    Adder(this, that, sum)
    sum
  }


  /**
   * Exercise 1
   */

  // Complete the operators "*" and "==="

  def *(that: Quantity): Quantity = ???
  
  def ===(that: Quantity): Quantity = ???



  /**
   * Exercise 2
   */

  // Complete the operators "square" and "sqrt"

  def square: Quantity = ???
  
  def sqrt: Quantity = ???


  override def toString(): String = value match {
   case None    => "  ?"
   case Some(v) => v.toString()
  }
}

/**
 * You can run the object ConstraintTests for testing your code. However,
 * we strongly suggest that you look at the test file "ConstraintSuite.scala"
 * and implement your unit test using the ScalaTest framework.
 */
object ConstraintTests extends App {
  def adderExample {
    val q1, q2, q3 = new Quantity
    Adder(q1, q2, q3)
    Probe("q1", q1)
    Probe("q2", q2)
    Probe("q3", q3)

    q1.setValue(20)
    q3.setValue(22)
    q1.forgetValue
    q3.forgetValue
  }

  adderExample
}
